import warnings
import numpy as np
import pandas as pd
from plotnine import *
from sklearn.preprocessing import scale
from typing import Union, Optional, Tuple, Collection, Sequence, Iterable

from ._utilities import get_gene_values


def plot_gene(liger_object,
              gene,
              use_raw: bool = False,
              use_scaled: bool = False,
              scale_by='dataset',
              log2scale: Optional[bool] = None,
              methylation_indices=None,
              plot_by='dataset',
              set_dr_lims=False,
              pt_size=0.1,
              min_clip=None,
              max_clip=None,
              clip_absolute=False,
              points_only=False,
              option='plasma_r',
              cols_use=None,
              zero_color='#F5F5F5',
              axis_labels=None,
              do_legend=True,
              return_plots=False,
              keep_scale=False,
              legend_text_size=12):
    """Plot gene expression on dimensional reduction (t-SNE) coordinates

    Generates plot of dimensional reduction coordinates (default t-SNE) colored by expression of
    specified gene. Data can be scaled by dataset or selected feature column from cell_data (or across
    all cells). Data plots can be split by feature.

    Parameters
    ----------
    liger_object : TYPE
        Should call runTSNE before calling.
    gene
        Gene for which to plot expression.
    use_raw : TYPE, optional
        DESCRIPTION. The default is False.
    use_scaled : TYPE, optional
        DESCRIPTION. The default is False.
    scale_by : TYPE, optional
        DESCRIPTION. The default is 'dataset'.
    log2scale : TYPE, optional
        DESCRIPTION. The default is None.
    methylation_indices : TYPE, optional
        DESCRIPTION. The default is None.
    plot_by : TYPE, optional
        DESCRIPTION. The default is 'dataset'.
    set_dr_lims : TYPE, optional
        DESCRIPTION. The default is False.
    pt_size : TYPE, optional
        DESCRIPTION. The default is 0.1.
    min_clip : TYPE, optional
        DESCRIPTION. The default is None.
    max_clip : TYPE, optional
        DESCRIPTION. The default is None.
    clip_absolute : TYPE, optional
        DESCRIPTION. The default is False.
    points_only : TYPE, optional
        DESCRIPTION. The default is False.
    option : TYPE, optional
        DESCRIPTION. The default is 'plasma'.
    cols_use : TYPE, optional
        DESCRIPTION. The default is None.
    zero_color : TYPE, optional
        DESCRIPTION. The default is '#F5F5F5'.
    axis_labels : TYPE, optional
        DESCRIPTION. The default is None.
    do_legend : TYPE, optional
        DESCRIPTION. The default is True.
    return_plots : TYPE, optional
        DESCRIPTION. The default is False.
    keep_scale :
        Maintain min/max color scale across all plots when using plot.by (default FALSE)

    Returns
    -------
    None.
    """
    if plot_by != scale_by and use_scaled:
        warnings.warn('Provided values for plot_by and scale_by do not match; results may not be very interpretable.')

    ### 1. Extract Gene Values
    if use_raw:
        if log2scale is None:
            log2scale = False

        # drop only outer level names
        gene_vals = get_gene_values(gene, data_use='raw', log2scale=log2scale)

    else:
        if log2scale is None:
            log2scale = True

        # rescale in case requested gene not highly variable
        if use_scaled:
            # check for feature
            if scale_by is not None and scale_by not in ['nUMI', 'nGene', 'dataset']:
                raise ValueError('Please select existing feature in cell_data to scale_by, or add it before calling.')

            gene_vals = get_gene_values(liger_object, gene, log2scale=log2scale)

            # set up dataframe with groups
            gene_df = pd.DataFrame({'gene': gene_vals}, dtype=np.float64)

            if scale_by is None:
                gene_df['scale_by'] = np.repeat('none', gene_vals.shape[0])
            else:
                gene_df['scale_by'] = liger_object.get_obs(scale_by, return_values=True)

            # scale by selected feature
            gene_df1 = gene_df.groupby('scale_by')['gene'].transform(scale, with_mean=False)
            gene_vals = gene_df1['gene']

            if log2scale:
                gene_vals = np.log2(10000 * gene_vals + 1)

        else:
            # using normalized data
            # indicate methylation indices here
            gene_vals = get_gene_values(liger_object, gene, methylation_indices=methylation_indices, log2scale=log2scale)

    if np.sum(gene_vals) != 0:
        gene_vals[gene_vals == 0] = np.nan


    ### 2. Build dataframe for plotting
    # extract min and max expression values for plot scaling if keep_scale = T
    if keep_scale:
        max_exp_val = np.nanmax(gene_vals)
        min_exp_val = np.nanmin(gene_vals)

    umap_coords = np.concatenate(liger_object.get_obsm('umap_coords'))
    dr_df = pd.DataFrame(umap_coords, columns=['tsne1', 'tsne2'])
    dr_df['gene'] = gene_vals

    # get dr limits for later
    lim1 = [dr_df['tsne1'].min(), dr_df['tsne1'].max()]
    lim2 = [dr_df['tsne2'].min(), dr_df['tsne2'].max()]

    if plot_by is not None:
        if plot_by not in ['nUMI', 'nGene', 'dataset']:
            raise ValueError('Please select existing feature in cell_data to plot_by, or add it before calling.')
        dr_df['plot_by'] = liger_object.get_obs(plot_by, return_values=True)
    else:
        dr_df['plot_by'] = np.repeat('none', gene_vals.shape[0])

    # expand clip values if only single provided
    num_levels = dr_df['plot_by'].nunique()
    if min_clip is None:
        min_clip = dict(zip(liger_object.sample_names, np.repeat(np.nan, num_levels)))

    if max_clip is None:
        max_clip = dict(zip(liger_object.sample_names, np.repeat(np.nan, num_levels)))

    ###!!!    #if min_clip is not None and
    ### 3. Create plot for each dataset
    p_list = {}
    for group_name, sub_df in dr_df.groupby('plot_by'):
        # maybe do quantile cutoff here
        if not clip_absolute:
            if not np.isnan(max_clip[group_name]):
                max_v = np.nanquantile(sub_df['gene'], q=max_clip[group_name])
            else:
                max_v = 1000
            if not np.isnan(min_clip[group_name]):
                min_v = np.nanquantile(sub_df['gene'], q=min_clip[group_name])
            else:
                min_v = 0
        else:
            max_v = max_clip[group_name]
            min_v = min_clip[group_name]

        sub_df.loc[(sub_df['gene'] > max_v) & (sub_df['gene'].notna()), 'gene'] = max_v
        sub_df.loc[(sub_df['gene'] < min_v) & (sub_df['gene'].notna()), 'gene'] = min_v

        ggp = (ggplot(data=sub_df, mapping=aes(x='tsne1', y='tsne2', color='gene')) +
               geom_point(size=pt_size) +
               labs(colour=gene))

        if cols_use is not None:
            if keep_scale:
                ggp = ggp + scale_color_gradientn(colors=cols_use, na_value=zero_color,
                                                  limits=[min_exp_val, max_exp_val])
            else:
                ggp = ggp + scale_color_gradientn(colors=cols_use, na_value=zero_color)
        else:
            if keep_scale:
                ggp = ggp + scale_color_cmap(cmap_name=option, na_value=zero_color, limits=[min_exp_val, max_exp_val])
            else:
                ggp = ggp + scale_color_cmap(cmap_name=option, na_value=zero_color)

        if set_dr_lims:
            ggp = ggp + xlim(lim1) + ylim(lim2)

        if plot_by is not None:
            base = sub_df['plot_by'].iloc[0]
        else:
            base = ''

        ggp = ggp + ggtitle(base)

        if axis_labels is not None:
            ggp = ggp + xlab(axis_labels[0]) + ylab(axis_labels[1])

        if not do_legend:
            ggp = ggp + theme(legend_position='none')

        if points_only:
            ggp = ggp + theme(axis_line=element_blank(), axis_text_x=element_blank(),
                              axis_text_y=element_blank(), axis_ticks=element_blank(),
                              axis_title_x=element_blank(), axis_title_y=element_blank(),
                              legend_position='none', panel_background=element_blank(),
                              panel_border=element_blank(), panel_grid_major=element_blank(),
                              panel_grid_minor=element_blank(), plot_background=element_blank(),
                              plot_title=element_blank())
        p_list[sub_df['plot_by'].iloc[0]] = ggp + theme_classic(legend_text_size)

    # if plot_by == 'dataset':
    #    p_list = p_list[]

    if return_plots:
        return p_list
    else:
        for plot in p_list:
            plot.draw()
        return None


def plot_gene_dict(liger_object, gene_dict):
    return None


def plot_gene_spatial(liger_object,
                      gene,
                      use_raw: bool = False,
                      use_scaled: bool = False,
                      scale_by='dataset',
                      log2scale: Optional[float] = None,
                      methylation_indices=None,
                      plot_by='dataset',
                      set_dr_lims=False,
                      pt_size=0.1,
                      min_clip=None,
                      max_clip=None,
                      clip_absolute=False,
                      points_only=False,
                      option='plasma_r',
                      cols_use=None,
                      zero_color='#F5F5F5',
                      axis_labels=None,
                      do_legend=True,
                      return_plots=False,
                      keep_scale=False):
    if plot_by != scale_by and use_scaled:
        warnings.warn('Provided values for plot_by and scale_by do not match; results may not be very interpretable.')

    ### 1. Extract Gene Values
    if use_raw:
        if log2scale is None:
            log2scale = False

        # drop only outer level names
        gene_vals = get_gene_values(gene, data_use='raw', log2scale=log2scale)

    else:
        if log2scale is None:
            log2scale = True

        # rescale in case requested gene not highly variable
        if use_scaled:
            # check for feature
            if scale_by is not None and scale_by not in ['nUMI', 'nGene', 'dataset']:
                raise ValueError('Please select existing feature in cell_data to scale_by, or add it before calling.')

            gene_vals = get_gene_values(liger_object, gene, log2scale=log2scale)

            # set up dataframe with groups
            gene_df = pd.DataFrame({'gene': gene_vals}, dtype=np.float64)

            if scale_by is None:
                gene_df['scale_by'] = np.repeat('none', gene_vals.shape[0])
            else:
                gene_df['scale_by'] = liger_object.get_obs(scale_by, return_values=True)

            # scale by selected feature
            gene_df1 = gene_df.groupby('scale_by')['gene'].transform(scale, with_mean=False)
            gene_vals = gene_df1['gene']

            if log2scale:
                gene_vals = np.log2(10000 * gene_vals + 1)

        else:
            # using normalized data
            # indicate methylation indices here
            gene_vals = get_gene_values(liger_object, gene, methylation_indices=methylation_indices, log2scale=log2scale)

    if np.sum(gene_vals) != 0:
        gene_vals[gene_vals == 0] = np.nan

    figs = []
    for adata in liger_object.adata_list:
        figs = _plot_spatial_adata(adata, image_key)

    if return_plots:
        return figs
    else:
        return None


import numpy as np
import pandas as pd

from scipy.sparse import vstack
from anndata import AnnData


"""
The pyliger main class
"""
class Liger(object):
    """Main LIGER class
    
    The liger object is created from two or more single cell datasets. To construct a
    liger object, the user needs to provide at least two expression (or another
    single-cell modality) matrices. The class serves as a container for results generated
    from  data preprocessing, integrative analysis, and visualization.
        
    Attributes:
        adata_list(list):
            List of AnnData objects, one per experiment/dataset (cells by genes)
            In each AnnData objects, main matrix stores raw data and two addtional
            layers store normalized and scaled data with keys 'norm_data' and
            'scale_data' respectively.
            H(matrix): 
            Cell loading factors (one matrix per dataset, dimensions cells by k)
            W(matrix): 
            Shared gene loading factors (k by genes)
            V(matrix): 
            Dataset-specific gene loading factors (one matrix per dataset, dimensions k by genes)
        cell_data(pd dataframe): 
            Dataframe of cell attributes across all datasets (nrows equal to total number
            cells across all datasets)
        var_genes(list): 
            Subset of informative genes shared across datasets to be used in matrix
            factorization
        H_norm(pd dataframe): 
            Normalized cell loading factors (cells across all datasets combined into single
            matrix)
        clusters(pd dataframe): 
            Joint cluster assignments for cells
        tsne_coords(): 
            Matrix of 2D coordinates obtained from running t-SNE on H_norm or H matrices
        alignment_clusters(): 
            Initial joint cluster assignments from shared factor alignment
        snf(list): 
            List of values associated with shared nearest factor matrix for use in clustering and
            alignment (out_summary contains edge weight information between cell combinations)
        agg_data(list): 
            Data aggregated within clusters
        parameters(list): 
            List of parameters used throughout analysis
        version(): 
            Version of package used to create object
    """
    
    __slots__ = ('adata_list', 'cell_data', 'var_genes', 'tsne_coords',
                 'alignment_clusters', 'agg_data', 'parameters',
                 'snf', 'version')
    
    def __init__(self, adata_list=[]):
        self.adata_list = adata_list

    #@property
    #def adata_list(self):
    #    return self._adata_list
    #@adata_list.setter
    #def adata_list(self):

    @property
    def num_samples(self):
        return len(self.adata_list)

    @property
    def num_var_genes(self):
        return len(self.var_genes)

    @property
    def sample_names(self):
        return [adata.uns['sample_name'] for adata in self.adata_list]

    @property
    def H(self):
        """num_cells x k"""
        return [adata.obsm['H'] for adata in self.adata_list]

    @property
    def V(self):
        """num_genes x k"""
        #return [adata.varm['V'][adata.uns['var_gene_idx'], :] for adata in self.adata_list]
        return [adata.varm['V'] for adata in self.adata_list]

    @property
    def W(self):
        """W is the shared component. Only return one is enough."""
        return self.adata_list[0].varm['W']

    def show(self):
        print("An object of class liger with {} datasets and {} total cells.".format(self.num_samples),len(self.cell_data))

    def add_adata(self, new_arrive):
        if isinstance(new_arrive, list):
            for adata in new_arrive:
                if isinstance(new_arrive, AnnData):
                    self.pending_list.append(adata)
                else:
                    print('Invalid input. Input must be list of AnnData object')
        elif isinstance(new_arrive, AnnData):
            self.pending_list.append(new_arrive)
        else:
            print('Invalid input. Input must be AnnData object')

    def save_raw(self):
        for i in range(self.num_samples):
            self.adata_list[i].raw = self.adata_list[i]

    def find_dataset_idx(self, dataset_name):
        for i in range(self.num_samples):
            if dataset_name == self.adata_list[i].uns['sample_name']:
                return i
            else:
                continue
        return 'Dataset does not exist'

    def get_data(self, set_name, dataset_use='all', combine=False, use_var=False):
        """"""
        if dataset_use == 'all':
            if set_name == 'raw':
                data = [adata.X for adata in self.adata_list]
            else:
                data = [adata.layers[set_name] for adata in self.adata_list]

            if combine:
                data = vstack(data)

        else:
            if set_name == 'raw':
                data = self.adata_list[dataset_use].X
            else:
                data = self.adata_list[dataset_use].layers[set_name]

        return data

    def get_obs(self, obs_name, return_values=False):
        obs_values = pd.concat([adata.obs[obs_name] for adata in self.adata_list])
        
        if return_values:
            return obs_values.values
        else:
            return obs_values

    def return_H(self, dataset_use='all'):
        H_list = []
        for adata in self.adata_list:
            if dataset_use == 'all':
                H_list.append(adata.obsm['H'])
            elif dataset_use == adata.uns['sample_name']:
                H_list.append(adata.obsm['H'])
            else:
                continue
        return H_list

    def return_raw(self, dataset_use='all'):
        for adata in self.adata_list:
            if dataset_use == 'all':
                yield adata.X
            elif dataset_use == adata.uns['sample_name']:
                yield adata.X

    def get_varm(self, var_name, dataset_use='all'):
        """

        Args:
            var_name:
            dataset_use:

        Returns:

        """
        if var_name == 'W':
            adata = self.adata_list[0]
            var_values = adata.varm[var_name][adata.uns['var_gene_idx'], :]
        else:
            if dataset_use == 'all':
                var_values = np.concatenate([adata.varm[var_name][adata.uns['var_gene_idx'], :] for adata in self.adata_list])
            else:
                adata = self.adata_list[dataset_use]
                var_values = adata.varm[var_name][adata.uns['var_gene_idx'], :]

        return var_values

    def get_gene_values(self, gene, use_cols=False, methylation_indices=None, log2scale=False,
                        scale_factor=10000):
        """"""

        if methylation_indices is None:
            methylation_indices = []

        gene_vals_total = []
        for idx, adata in enumerate(self.adata_list):
            gene_names = self.adata_list[idx].var.index
            if gene in gene_names:
                gene_vals = np.ravel(adata[:, gene].layers['norm_data'].toarray())
            else:
                gene_vals = np.zeros(adata.shape[0], dtype=np.int)

            if log2scale and idx not in methylation_indices:
                gene_vals = np.log2(gene_vals * scale_factor + 1)

            gene_vals_total.append(gene_vals)

        return np.concatenate(gene_vals_total)

    def save_obsm(self, obsm_value, obsm_name, dataset_use='all'):
        if dataset_use == 'all':
            dataset_use = list(range(self.num_samples))

        idx = 0
        for i in dataset_use:
            self.adata_list[i].obsm[obsm_name] = obsm_value[idx:(idx+self.adata_list[i].shape[0])]
            idx += self.adata_list[i].shape[0]

        return None

    def get_obsm(self, obsm_name, dataset_use='all'):
        """
        """
        if dataset_use == 'all':
            dataset_use = list(range(self.num_samples))

        obsm_values = []
        for i in dataset_use:
            obsm_values.append(self.adata_list[i].obsm[obsm_name])

        return obsm_values


    def save(self):
        """

        :return:
        """
        pass


    def load(self):
        pass



import matplotlib

matplotlib.use('Agg')
import sys
import logging
from datetime import date
from collections import defaultdict
from collections import OrderedDict
from sklearn.feature_extraction import DictVectorizer
import numpy as np
from sklearn.manifold import MDS
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from math import log
from math import log10
import matplotlib.patches as mpatches
import os
import argparse
from adjustText import adjust_text  ## Python package which rearranges plot labels to ensure minimal overlap
from matplotlib.colors import ListedColormap


def plot_go_term(input_goea,
                 output_dir_path,
                 max_clusters=50,
                 ontology='all',
                 representatives=None,
                 similarity_cutoff=0.5,
                 max_pvalue=99999.99999,
                 sort_by='pval',
                 sum_user=False,
                 top_level=None,
                 name_changes=None,
                 opacity=1.0,
                 colours='log10-pval',
                 palette='plasma',
                 random_state=1,
                 size='members',
                 size_range='medium',
                 cluster_labels='numbered',
                 xlabel='Semantic space X',
                 ylabel='Semantic space Y',
                 colour_label=None,
                 legend_position='center',
                 description_limit=35,
                 font_size='medium',
                 legend_columns='double',
                 legend='description',
                 max_labels=10,
                 dpi=400,
                 file_type='png',
                 outfile_appendix=None,
                 title_size='medium',
                 title=None):
    """
    Wrapper function to run GO-Figure!
    Args:
        input_dict(dict):
            Input file in tabular format with the columns: GO ID + P-value for standard input,
            GO ID + P-Value + Significant for standard-plus input, TopGO output as an input,
            or GoStats output as an input. Can use # or ! or %% to comment out lines.
        output_dir_path(str):
            Output directory
        max_clusters(int): optional, (default 50.)
            Maximum amount of clusters to plot (integer value).
        ontology(str): optional, (default 'all')
            Which ontology to use: biological process ('bpo'), molecular function ('mfo'),
            cellular component ('cco'), or all ontologies ('all').
            choices=['bpo', 'mfo', 'cco', 'all']
        representatives(str): optional, (default None)
            A list of GO terms that have priority for being a representative of a cluster.
            I.e. if one term in a cluster has priority, that term will always be the representative.
            If two terms in a cluster have priority, only those two will be considered.
            Input is a list of GO terms separated by a comma, such as 'GO:0000001,GO:0000002'.
        similarity_cutoff(float): optional, (default 0.5)
            Similarity cutoff to be used between GO terms, a value between 0 and 1.
        max_pvalue(float): optional, (default 99999.99999)
            Maximum p-value to consider (floating value).
        sort_by(str): optional, (default 'pval')
            Which values to use for sorting the clusters: 'pval' (p-value) or 'user' (user-defined value)
            or 'user-descending' (user-defined value descending).
            choices=['pval', 'user', 'user-descending']
        sum_user(bool): optional, (default False)
            To sum the user-defined values (column 3) for each member of a cluster. Either 'True' or 'False'.
            choices=[True, False]
        top_level(str): optional, (default None)
            Set top level GO terms for clusters as given by the GO DAG (see https://www.ebi.ac.uk/QuickGO).
            Top level GO terms have to be given separated by comma's, without spaces. E.g.: 'GO:000001,GO:000008'.
        name_changes(str): optional, (default None)
            A list of GO terms that will be used as representative of a cluster specifically for naming purposes,
            but not for internal calculations. This is opposed to the'--representatives option, which provides
            GO terms to be used as representatives of a cluster both internally and externally. This specific
            option allows for the renaming of clusters without recalculating the clusters when there is a need
            to reproduce the original figure. Input is a list of GO terms separated by a comma,
            such as 'GO:0000001,GO:0000002'.
        opacity(float): optional, (default 1.0)
            Set opacity for the clusters, floating point from 0 to 1.
        colours(str): optional, (default 'log10-pval')
            Color GO clusters based on p-value ('pval'), log10 p-value ('log10-pval'), number of GO terms that
            are a member of the cluster ('members'), frequency of the GO term in the GOA database ('frequency'),
            a unique colour per cluster ('unique'), or a user defined value ('user'). Default = 'log10-pval'.
            choices=['pval', 'log10-pval', 'user', 'members', 'frequency', 'unique']
        palette(str): optional, (default 'plasma')
            Set to any color brewer palette available for MatPlotLib (https://matplotlib.org/3.1.1/tutorials/colors/colormaps.html).
        random_state(int): optional, (default 1)
            Set random state for the calculation of the X and Y label. Useful if you want the figures to be exactly the same.
            Needs to be an integer or None. Specifying 'None' will create a different orientation of the plot every time.
        size(str): optional, (default 'members')
            Size of GO clusters based on how many GO terms are a member of the cluster ('members'), frequency in GOA
            database ('frequency'), p-value where smaller = larger size ('pval'), the user defined value ('user'), or
            a fixed integer for every cluster.
        size_range(str): optional, (default 'medium')
            Set cluster size range to 'small', 'medium', or 'big'.
            choices=['small', 'medium', 'big']
        cluster_labels(str): optional, (default 'numbered')
            Label clusters numbered based on the sorting option ('numbered'), based on the representative GO term ('go'),
            based on the representative GO term with arrows ('go-arrows') based on the representative GO term
            name ('description'), or based on the representative GO term name with arrows ('description-numbered').
            choices=['numbered', 'go', 'go-arrows', 'description', 'description-numbered']
        xlabel(str): optional, (default 'Semantic space X')
            X-axis label.
        ylabel(str): optional, (default 'Semantic space Y')
            Y-axis label
        colour_label(str): optional, (default 'Semantic space Y')
            Colour bar label. Default is the description of the input used for --colours
        legend_position(str): optional, (default 'center')
            Position the legend at the bottom left ('left'), bottom right ('right'), or bottom center ('center').
            choices=['left', 'right', 'center']
        description_limit(int): optional, (default 35)
            Integer character limit of GO term descriptions in the legend.
        font_size(str): optional, (default 'medium')
            Font size of the legend. 'xx-small', 'x-small', 'small', 'medium', 'large' , 'x-large', or 'xx-large'.
            choices=['xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large']
        legend_columns(str): optional, (default 'double')
            Legend as a single ('single') column or double ('double') column.
            choices=['single', 'double', 'triple']
        legend(str): optional, (default 'description')
            Option to show GO terms and descriptions in the legend ('full'), GO term only 'go', description
            only ('description'), or no legend ('none').
            choices=['full', 'go', 'description', 'none']
        max_labels(int): optional, (default 10)
            Maximum labels to display in the legend.
        dpi(int): optional, (default 400)
            Set DPI for outfiles.
        file_type(str): optional, (default 'png')
            Image file type. 'png', 'pdf', 'svg', or 'tiff'
            choices=['png', 'pdf', 'tiff', 'svg']
        outfile_appendix(str): optional, (default None)
            What to add to the outfiles after 'biological_process', 'molecular_function', or 'cellular_component'.
            By default it will add the value given for --outdir, replacing '/' with '_'.
        title_size(str): optional, (default 'medium')
            Set title size. 'xx-small', 'x-small', 'small', 'medium', 'large' , 'x-large', or 'xx-large'.
            choices=['xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large']
        title(str): optional, (default None)
            Set figure title. Has to be between single or double quotation marks.

    Return:
    """
    if not os.path.isdir(output_dir_path):
        os.mkdir(output_dir_path)

    if outfile_appendix is None:
        outfile_appendix = ''

    zlabel = colour_label

    ## Initialize logger instances
    logger, logger_std = create_logger(output_dir_path, outfile_appendix)

    ## Convert representatives input to a list of GO terms
    if representatives == None:
        representatives = []
    else:
        representatives = representatives.split(',')

    ## Check if representatives are valid GO terms
    for priority in representatives:
        if not priority.startswith('GO:') or not len(priority) == 10:
            logger.critical(
                'gofigure.py: error: argument -r/--representatives contains a GO term that is not in valid GO term format: ' + priority)
            exit()

    try:
        random_state = int(random_state)
    except:
        if random_state.lower() == 'none':
            random_state = None
        else:
            logger.critical("gofigure.py: error: argument -rs/--random_state: invalid choice: '" + str(
                random_state) + "' (requires an integer or 'None'")
            exit()

    ## Opacity is a number from 0 to 1
    if opacity < 0 or opacity > 1:
        logger.critical("gofigure.py: error: argument -o/--opacity: invalid choice: '" + str(
            opacity) + "' (requires a floating point from 0 to 1)")
        exit()
    if not size in ['user', 'pval', 'frequency', 'members']:
        try:
            size = int(size)
        except:
            logger.critical(
                "gofigure.py: error: argument -s/--size: invalid choice: '" + size + "' (choose from 'members', 'frequency', 'pval', 'significant', or an integer)")
            exit()

    ## Similarity cutoff is a number from 0 to 1
    if similarity_cutoff < 0 or similarity_cutoff > 1:
        logger.critical("gofigure.py: error: argument -c/--colours: invalid choice: '" + str(
            similarity_cutoff) + "' (choose any value ranging from 0 to 1)")
        exit()

    ## See if the given colour palette is valid
    try:
        cmap = ListedColormap(sns.color_palette(palette))
    except:
        logger.critical(
            "gofigure.py: error: argument -p/--palette: invalid choice: '" + palette + "' (for example, pick one of https://matplotlib.org/3.1.1/tutorials/colors/colormaps.html)")
        exit()


    logger.info('GO-Figure! version: 1.0\n')
    logger.info('Date run: ' + str(date.today()) + '\n')
    logger.info('Command line input: ' + str(sys.argv) + '\n')

    logger_std.info('Starting calculations')

    input_dict = process_input(input_goea, logger)


    ## code need an update
    scriptPath = '.'
    ic_dict, frequency_dict = read_IC(scriptPath, logger)
    description_dict, namespace_dict, obsolete_dict, alt_dict = read_OBO(scriptPath, logger)
    parents_dict, children_dict = read_parents_children(scriptPath, logger)

    if ontology == 'all':  ## Which ontologies to loop over
        ontology_list = ['biological_process', 'molecular_function', 'cellular_component']
    elif ontology == 'bpo':
        ontology = 'biological_process'
        ontology_list = [ontology]
    elif ontology == 'mfo':
        ontology = 'molecular_function'
        ontology_list = [ontology]
    elif ontology == 'cco':
        ontology = 'cellular_component'
        ontology_list = [ontology]

    for ontology in ontology_list:
        logger_std.info('\nCalculating ' + ontology.replace('_', ' '))
        logger.info('\nCalculating ' + ontology.replace('_', ' ') + '\n')
        go_dict = create_GO_dict(input_dict, ontology, namespace_dict, ic_dict, frequency_dict, max_pvalue,
                                 alt_dict, obsolete_dict, logger)
        semsim_dict = create_clusters(go_dict, parents_dict, children_dict, ic_dict, similarity_cutoff, representatives, top_level)
        goset = set()
        for key, values in semsim_dict.items():
            goset.add(key)
            for value in values:
                goset.add(value[0])
        cluster_dict = create_clusterdict(semsim_dict, description_dict)
        if not len(semsim_dict) == 0:
            df, semsim_dict = create_df(cluster_dict, go_dict, description_dict, colours, size, description_limit,
                                        legend, cluster_labels, random_state, semsim_dict, name_changes, parents_dict, ic_dict, sum_user, sort_by)
            scatterplot(df, output_dir_path, ontology, max_labels, palette, colours, size, legend, legend_position,
                        max_clusters, legend_columns, cluster_labels, opacity, size_range, font_size, title, title_size,
                        dpi, file_type, outfile_appendix, xlabel, ylabel, zlabel, logger_std)
            output_table(df, output_dir_path, ontology, outfile_appendix)
            create_cluster_table(semsim_dict, description_dict, output_dir_path, ontology, outfile_appendix)
        else:
            logger_std.info('No GO terms for ontology found: ' + ontology)
            logger.info('No GO terms for ontology found: ' + ontology + '\n')

    logger_std.info("Finished!")

    return None


##
## MaartenViz: blablablabla
##
## Authors: Maarten JMF Reijnders and Robert M Waterhouse
##
## Contact: maarten.reijnders@unil.ch or robert.waterhouse@unil.ch
##

def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn


## Creating logfiles. Logger_std is for writing to the console, logger is for writing to a file.
def create_logger(output_dir_path, appendix):
    logger = logging.getLogger('maartenviz')
    logger_std = logging.getLogger('maartenviz_std')
    logger.setLevel(logging.INFO)
    logger_std.setLevel(logging.INFO)

    if appendix == '':
        error_file_handler = logging.FileHandler(
            output_dir_path + '/logfile_' + output_dir_path.replace('/', '_') + '.txt', mode='w')
    else:
        error_file_handler = logging.FileHandler(output_dir_path + '/logfile_' + appendix + '.txt', mode='w')
    error_console_handler = logging.StreamHandler()
    error_console_handler.setLevel(logging.CRITICAL)
    logger.addHandler(error_file_handler)
    logger.addHandler(error_console_handler)

    console_handler = logging.StreamHandler()

    logger_std.addHandler(console_handler)
    logger_std.propagate = False
    logger.propagate = False
    return (logger, logger_std)


def process_input(input_goea, logger):
    """
    helper function to process the input GOEA results
    :param input_goea:
    :param logger:
    :return:
    """
    input_dict = defaultdict(list)
    go_set = set()
    line_count = 1

    for go in input_goea:
        term = go.GO
        pvalue = go.get_pvalue()
        if term in go_set:
            logger.error('Found duplicate GO term ID: ' + term + ' at line ' + str(
                line_count) + '. Please fix duplicates before continuing.')
            exit()
        go_set.add(term)
        input_dict[term] = [term, pvalue]
        line_count += 1

    return input_dict


## Instead of calculating the IC contents on the spot, read them from the IC file. This means less input file requirements and is slightly faster.  The file is created using scripts/ics.py.
def read_IC(scriptPath, logger):
    ic_dict = defaultdict(float)
    frequency_dict = defaultdict(float)
    for line in open(scriptPath + '/ic.tsv'):
        if not line.startswith('#'):
            go, ic, frequency = line.strip().split('\t')
            ic_dict[go] = float(ic)
            frequency_dict[go] = float(frequency)
        elif line.startswith('# GOA version used: '):
            logger.info('GOA version used to calculate IC and frequencies: ' + line.strip().split(': ')[1] + '\n')
    return (ic_dict, frequency_dict)


## Read the GO OBO file to match GO terms to their names and namespaces (description_dict and namespace_dict, respectively)
def read_OBO(scriptPath, logger):
    description_dict = defaultdict(str)
    namespace_dict = defaultdict(str)
    obsolete_dict = defaultdict(set)
    alt_dict = defaultdict(str)
    for line in open(scriptPath + '/go.obo'):
        if line.startswith('id: GO:'):
            go = line.strip().split('id: ')[1]
            obsolete_bool = False
        elif line.startswith('alt_id: GO:'):
            alt_id = line.strip().split('alt_id: ')[1]
            alt_dict[alt_id] = go
        elif line.startswith('name: '):
            name = line.strip().split('name: ')[1]
            description_dict[go] = name
        elif line.startswith('namespace: '):
            namespace = line.strip().split('namespace: ')[1]
            namespace_dict[go] = namespace
        elif line.lower().startswith('is_obsolete: true'):
            obsolete_bool = True
            del description_dict[go]
            del namespace_dict[go]
        elif line.startswith('consider: GO:') and obsolete_bool == True:
            consider_go = line.strip().split('consider: ')[1]
            obsolete_dict[go].add(consider_go)
        elif line.startswith('data-version: '):
            obo_version = line.strip().split('data-version: ')[1]
            logger.info('go.obo version used: ' + obo_version + '\n')
    return (description_dict, namespace_dict, obsolete_dict, alt_dict)


## Read child - parent relations, used for calculating semantic similarities
def read_parents_children(scriptPath, logger):
    parents_dict = defaultdict(list)
    children_dict = defaultdict(list)
    for line in open(scriptPath + '/relations_full.tsv'):
        if not line.startswith('#'):
            go, parent = line.strip().split('\t')
            parents_dict[go].append(parent)
            children_dict[parent].append(go)
        elif line.startswith('#go.obo version used: '):
            logger.info('go.obo version used to create GO relations: ' + line.strip().split(': ')[1] + '\n')
    return (parents_dict, children_dict)


## Calculates semantic similarities using Lin's formula
def calc_sem_sim(go1, go2, parents_dict, ic_dict):
    go1_parents = parents_dict[go1]
    go2_parents = parents_dict[go2]
    go1_IC = ic_dict[go1]
    go2_IC = ic_dict[go2]
    parent_IC = 0
    sem_sim = 0
    if go1 == go2:  ## If both GO terms are the same, semantic similarity = 1
        sem_sim = 1
    elif go1 in go2_parents:  ## If go1 is a parent of go2, the parent term used in the sem_sim equation is go1
        parent_IC = go1_IC
    elif go2 in go1_parents:  ## If go2 is a parent of go1, the parent term used in the sem_sim equation is go2
        parent_IC = go2_IC
    else:
        parent_IC = 0
        closest_parent = ''
        for parent in go1_parents:  ## If neither GO is a parent of the other, calculate common parent term with the highest IC content (meaning closest parent term)
            if parent in go2_parents:
                if ic_dict[parent] > parent_IC:
                    parent_IC = ic_dict[parent]
                    closest_parent = parent
    if not go1 == go2:
        if not go1_IC == 0 and not go2_IC == 0:  ## If one of the GO terms has no IC, something is wrong. Check if the GO term is available in the data files
            sem_sim = (2 * parent_IC) / (go1_IC + go2_IC)  ## Calculate semantic similarity
    return (sem_sim)


## Creates
def create_GO_dict(input_dict, namespace_input, namespace_dict, ic_dict, frequency_dict, max_pval,
                   alt_dict, obsolete_dict, logger):
    go_dict = defaultdict(list)
    conversion_bool = False

    for key, values in input_dict.items():
        go, pval = values
        pval = float(pval)
        if not go in namespace_dict:
            alt_dict_namespace = False
            if go in alt_dict:
                alt_go = go
                go = alt_dict[go]
                if namespace_dict[go] == namespace_input:
                    if conversion_bool == False:
                        logger.info(
                            "The following GO terms are converted or deleted because either they are an alt_id and not a primary id in the go.obo, because they are obsolete in the go.obo, or because they can't be found in the go.obo: \n\nOriginal GO\tNew GO(s)\tReason")
                        conversion_bool = True
                    logger.info(alt_go + '\t' + go + '\talt_id')
                    alt_dict_namespace == True
            elif not go in alt_dict or alt_dict_namespace == False:
                if conversion_bool == False:
                    logger.info(
                        "The following GO terms are converted or deleted because either they are an alt_id and not a primary id in the go.obo, because they are obsolete in the go.obo, or because they can't be found in the go.obo: \n\nOriginal GO\tNew GO(s)\tReason")
                    conversion_bool = True
                logger.info('\t' + go + '\tremoved')
        if namespace_dict[go] == namespace_input and max_pval >= pval:
            if go in obsolete_dict:
                if conversion_bool == False:
                    logger.info(
                        "The following GO terms are converted or deleted because either they are an alt_id and not a primary id in the go.obo, because they are obsolete in the go.obo, or because they can't be found in the go.obo: \n\nOriginal GO\tNew GO(s)\tReason")
                    conversion_bool = True
                for new_go in obsolete_dict[go]:
                    ic = float(ic_dict[new_go])
                    frequency = float(frequency_dict[new_go])
                    go_dict[new_go] = [pval, 0, ic, frequency]
                logger.info(go + '\t' + str(obsolete_dict[go]).replace('{', '').replace("'", '').replace('}',
                                                                                                         '') + '\tObsolete')
            else:
                if namespace_dict[go] == namespace_input and max_pval >= float(pval):
                    ic = float(ic_dict[go])
                    frequency = float(frequency_dict[go])
                    go_dict[go] = [pval, 0, ic, frequency]

    if 'GO:0008150' in go_dict:
        del go_dict['GO:0008150']
    if 'GO:0003674' in go_dict:
        del go_dict['GO:0003674']
    if 'GO:0005575' in go_dict:
        del go_dict['GO:0005575']
    return (go_dict)


## Creates clusters of GO terms, and chooses a representative term based on the workflow outlined below
def create_clusters(go_dict, parents_dict, children_dict, ic_dict, similarity_cutoff, priorities, top_level):
    go_dict = OrderedDict(go_dict.items())
    if priorities == None:
        priorities = []
    bool = True
    sem_sim_dict = defaultdict(list)
    prev_set = set()

    if top_level == None:
        top_level_terms = []
    else:
        top_level_terms = top_level.split(',')
    top_level_dict = defaultdict(float)
    for top_level_term in top_level_terms:
        if top_level_term in go_dict:
            pval, user, ic, frequency = go_dict[top_level_term]
            top_level_dict[top_level_term] = ic
    while bool == True:
        bool2 = True
        golist = list(OrderedDict(go_dict))
        for go in golist:  ## Loop over every GO term
            if go in go_dict:
                pval, user, ic, frequency = go_dict[go]
                highest_sem_sim = 0
                highest_sem_sim_GO = ''
                highest_sem_sim_pval = 100
                highest_sem_sim_user = 0
                highest_sem_sim_frequency = 1
                highest_sem_sim_IC = 0
                highest_top_level_ic = 0
                most_informative_top_level_term = ''
                for top_level_term in top_level_terms:
                    if go in children_dict[top_level_term]:
                        top_level_ic = top_level_dict[top_level_term]
                        if top_level_ic > highest_top_level_ic:
                            highest_top_level_ic = top_level_dict[top_level_term]
                            most_informative_top_level_term = top_level_term
                for go2 in golist:
                    go2_top_level_term = go2
                    if go in top_level_terms:
                        go_top_level_bool = True
                    else:
                        go_top_level_bool = False
                    go2_top_level_bool = False
                    if go2 in top_level_terms:
                        go2_top_level_bool = True
                    else:
                        for member in sem_sim_dict[go2]:
                            if member[0] in top_level_terms:
                                go2_top_level_bool = True
                                go2_top_level_term = member[0]
                    if not most_informative_top_level_term == '' and not go2_top_level_term == most_informative_top_level_term:
                        continue

                    # for go2,values2 in go_dict.items(): ## Loop over every GO term within the loop so they can be compared
                    if not go2 == go and go2 in go_dict:  ## Exclude the same GO term
                        if (go_top_level_bool == False and go2_top_level_bool == False) or (
                                go_top_level_bool == True and go2_top_level_bool == False and go2 in children_dict[
                            go]) or (go_top_level_bool == False and go2_top_level_bool == True and go in children_dict[
                            go2_top_level_term]):
                            pval2, user2, ic2, frequency2 = go_dict[go2]
                            sem_sim = calc_sem_sim(go, go2, parents_dict, ic_dict)
                            if sem_sim > highest_sem_sim:  ## Calculate the GO term which has the highest semantic similarity to the original GO term and adjust all values accordingly
                                highest_sem_sim = sem_sim
                                highest_sem_sim_GO = go2
                                highest_sem_sim_pval = pval2
                                highest_sem_sim_user = user2
                                highest_sem_sim_frequency = frequency2
                                highest_sem_sim_IC = ic2

                if highest_sem_sim >= similarity_cutoff:  ## If the highest semantic similarity is higher than the cutoff proceed with grouping the GO terms and choosing the representative
                    go_parents = parents_dict[go]
                    highest_sem_sim_GO_parents = parents_dict[highest_sem_sim_GO]
                    go1_bool = False
                    go2_bool = False
                    if go in priorities and highest_sem_sim_GO not in priorities:
                        sem_sim_dict[go].append([go, pval, user, ic, frequency])
                        sem_sim_dict[go].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go2_bool = True
                    elif highest_sem_sim_GO in priorities and not go in priorities:
                        sem_sim_dict[highest_sem_sim_GO].append([go, pval, user, ic, frequency])
                        sem_sim_dict[highest_sem_sim_GO].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go1_bool = True
                    elif frequency > 0.05 and not highest_sem_sim_frequency > 0.05:  ## If the frequency of the original GO term in the entire GOA database is higher than 0.05, exclude this GO term as a potential representative term
                        sem_sim_dict[highest_sem_sim_GO].append([go, pval, user, ic, frequency])
                        sem_sim_dict[highest_sem_sim_GO].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go1_bool = True
                    elif highest_sem_sim_frequency > 0.05 and not frequency > 0.05:  ## If the frequency of the second GO term in the entire GOA database is higher than 0.05, exclude this GO term as a potential representative term
                        sem_sim_dict[go].append([go, pval, user, ic, frequency])
                        sem_sim_dict[go].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go2_bool = True
                    elif highest_sem_sim_pval / pval < 0.5:  ## If the p-value of the original term is significantly higher than the 2nd term, the 2nd term is the representative
                        sem_sim_dict[highest_sem_sim_GO].append([go, pval, user, ic, frequency])
                        sem_sim_dict[highest_sem_sim_GO].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go1_bool = True
                    elif pval / highest_sem_sim_pval < 0.5:  ## If the p-value of the 2nd term is significantly higher than the original term, the original term is the representative
                        sem_sim_dict[go].append([go, pval, user, ic, frequency])
                        sem_sim_dict[go].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go2_bool = True
                    elif go in highest_sem_sim_GO_parents:  ## If the original GO term is a parent of the 2nd GO term, the 2nd GO term is the representative
                        sem_sim_dict[highest_sem_sim_GO].append([go, pval, user, ic, frequency])
                        sem_sim_dict[highest_sem_sim_GO].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go1_bool = True
                    elif highest_sem_sim_GO in go_parents:  ## If the 2nd GO term is a parent of the original GO term, the original GO term is the representative
                        sem_sim_dict[go].append([go, pval, user, ic, frequency])
                        sem_sim_dict[go].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go2_bool = True
                    else:  ## If none of the above reaches to a representative, the original GO term is the representative
                        sem_sim_dict[go].append([go, pval, user, ic, frequency])
                        sem_sim_dict[go].append(
                            [highest_sem_sim_GO, highest_sem_sim_pval, highest_sem_sim_user, highest_sem_sim_IC,
                             highest_sem_sim_frequency])
                        go2_bool = True
                    if go1_bool == True:  ## If the original term is not chosen as a representative and already has an entry in the sem_sim_dict, add its values to the 2nd GO term in the sem_sim_dict and delete the old entry
                        if go in sem_sim_dict:
                            for val in sem_sim_dict[go]:
                                if not val in sem_sim_dict[highest_sem_sim_GO]:
                                    sem_sim_dict[highest_sem_sim_GO].append(val)
                            del sem_sim_dict[go]
                        del go_dict[go]

                    if go2_bool == True:
                        if highest_sem_sim_GO in sem_sim_dict:
                            for val in sem_sim_dict[highest_sem_sim_GO]:
                                if not val in sem_sim_dict[go]:
                                    sem_sim_dict[go].append(val)
                            del sem_sim_dict[highest_sem_sim_GO]
                        del go_dict[highest_sem_sim_GO]

                elif not go in sem_sim_dict:  ## If no go term is found with a semantic similarity >= cutoff, make a new entry in the sem_sim_dict
                    sem_sim_dict[go].append([go, pval, user, ic, frequency])
        for go in golist:  ## This loop makes sure the above workflow is looped until no changes are made, by checking if the go_dict keys are present as a key in sem_sim_dict. Any key not present in sem_sim_dict after every loop is deleted. Once all go_dict keys == sem_sim_dict keys, bool2 == true and therefore bool == false and while loop ends
            if not go in sem_sim_dict:
                bool2 = False
                if go in go_dict:
                    del go_dict[go]
        if bool2 == True:
            bool = False
    return (sem_sim_dict)


## Creates a cluster dict containing just the bare information needed to create the dataframe for the scatterplot
def create_clusterdict(sem_sim_dict, description_dict):
    cluster_dict = defaultdict(list)
    for key, values in sem_sim_dict.items():
        dup_set = set()
        for value in values:
            go = value[0]
            user = value[2]
            if not go in dup_set:
                dup_set.add(go)
                name = description_dict[go]
                key_name = description_dict[key]
                cluster_dict[key].append([go, float(user)])
    return (cluster_dict)


## Creates the data frame for the scatterplot
def create_df(cluster_dict, go_dict, description_dict, colours, size, description_limit, legend, cluster_labels,
              random_state_input, semsim_dict, name_changes, parents_dict, ic_dict, sum_user, sort_by):
    if not random_state_input == None:
        random_state_input = int(random_state_input)
    order = list(cluster_dict.keys())  ## Keep static order of df
    sem_array = np.zeros((len(order), len(order)))

    for i in range(0,
                   len(order)):  ## Create a sem_array containing semantic similarities between each representative of each cluster
        go1 = order[i]
        for k in range(0, len(order)):
            go2 = order[k]
            sem = calc_sem_sim(go1, go2, parents_dict, ic_dict)
            sem_array[i, k] = sem

    embedding = MDS(n_components=2, dissimilarity='euclidean',
                    random_state=random_state_input)  ## Two dimensional array based on semantic similarities between all representative GO terms
    if len(sem_array) == 0:
        return (pd.DataFrame())

    sem_array_transformed = embedding.fit_transform(sem_array)
    df = pd.DataFrame(data=sem_array_transformed, index=order,
                      columns=['x', 'y'])  ## Create the dataframe starting with the x and y semantic similarities

    names = df.index.values

    frequency_list = []  ## Frequency of GO in GOA database
    colour_criterium_list = []  ## Values which will be used to colour clusters
    pval_list = []  ## Pvalues
    description_list = []  ## Names of GO terms
    count_list = []  ## Counts of GO terms in cluster
    legend_list = []  ## DF holding the legend descriptions
    size_criterium_list = []  ## Values which will be used to size the clusters in the scatterplot
    go_list = []  ## List of GO terms part of the cluster
    count = 1  ## Index number
    user_defined_values_list = []  ## User defined value

    if name_changes == None:
        name_changes_list = []
    else:
        name_changes_list = name_changes.split(',')

    for go in names:  ## Loop over each cluster and add values to above lists
        frequency = go_dict[go][2]
        if frequency == 0:  ## If it somehow bugs out, make 0.001 so it wont throw error
            frequency = 0.001
        colours = colours.lower()
        colour_criterium = log10(float(go_dict[go][0]))
        pval = float(go_dict[go][0])
        if sum_user == False:
            user = go_dict[go][1]
        pval_list.append(pval)
        goterms = cluster_dict[go]
        tmpGOList = []
        new_go = ''
        new_description = ''
        for goterm in goterms:  ## Add GO terms and user defined values
            tmpGOList.append(goterm[0])
            if goterm[0] in name_changes_list:
                new_go = goterm[0]
                new_description = description_dict[goterm[0]]
        if sum_user == False:
            user = go_dict[go][1]
        elif sum_user == True:
            user = 0
            for goterm in goterms:
                user += go_dict[goterm[0]][1]
        go_list.append(tmpGOList)
        user_defined_values_list.append(user)
        if colours == 'pval':  ## Decide which criteria to use for colouring of clusters
            colour_criterium = float(go_dict[go][0])
        elif colours == 'log10-pval':
            colour_criterium = log10(float(go_dict[go][0]))
        elif colours == 'user':
            colour_criterium = user
        elif colours == 'members':
            colour_criterium = 1 + len(goterms)
        elif colours == 'frequency':
            colour_criterium = -log(frequency)
        if size == 'user':  ## Decide which criteria to use for sizeing of clusters
            size_criterium = user
        elif size == 'pval':
            size_criterium = -log10(pval)
        elif size == 'frequency':
            size_criterium = log(frequency)
        elif size == 'members':
            size_criterium = log(1 + len(goterms))
        else:
            size_criterium = size
        frequency_list.append(log(frequency))
        colour_criterium_list.append(colour_criterium)
        size_criterium_list.append(size_criterium)
        description = description_dict[go]
        if not new_description == '':
            description = new_description
        if not new_go == '':
            names = np.where(names == go, new_go, names)
            semsim_dict[new_go] = semsim_dict.pop(go)
            go = new_go
        if len(description) > description_limit:  ## If a description exceeds the description character limit, truncate it
            description = description[0:description_limit] + '...'
        description_list.append(description)
        count_list.append(count)
        if legend.lower() == 'description':  ## Decide how to fill the legend
            legend_list.append('. ' + description)
        elif legend.lower() == 'go':
            legend_list.append('. ' + go)
        elif legend.lower() == 'full':
            legend_list.append('. ' + go + ' ' + description)
        elif legend.lower() == 'none':
            legend_list.append('. ')
        elif legend.lower() == 'exhaustive':
            legend_list.append('. ' + go + ' ' + description)
        count += 1
    df['frequency'] = frequency_list
    df['colour'] = colour_criterium_list
    df['size'] = size_criterium_list
    df['pval'] = pval_list
    df['description'] = description_list
    df['legend'] = legend_list
    df['members'] = go_list
    df['representative'] = names
    df['user'] = user_defined_values_list
    if sort_by == 'user-descending':
        df = df.sort_values(by='user', ascending=False)
    else:
        df = df.sort_values(by=sort_by)  ## Sorts the DF based on P-value (lowest first)

    df['dotCount'] = count_list
    df = df.set_index('dotCount')
    for index, row in df.iterrows():
        if legend == 'exhaustive':  ## Add member GO terms to exhaustive legend
            count = 0
            legend_description = ''
            for go in row['members']:
                description = description_dict[go]
                if len(description) > description_limit:
                    description = description[0:description_limit] + '...'
                if count == 0:
                    if cluster_labels.lower() == 'numbered':
                        legend_description = str(index) + '. ' + go + ' ' + description
                    elif cluster_labels.lower() == 'go':
                        legend_description = str(index) + '. ' + go + ' ' + description
                    count += 1
                else:
                    legend_description = legend_description + '\n    ' + go + ' ' + description
            df.loc[index, 'legend'] = legend_description
        else:  ## Else, fill legend normally
            if cluster_labels.lower() == 'numbered':
                df.loc[index, 'legend'] = str(index) + row['legend']
            elif cluster_labels.lower() == 'go' or cluster_labels.lower() == 'go-arrows' or cluster_labels.lower() == 'description-numbered' or cluster_labels.lower() == 'description':
                df.loc[index, 'legend'] = str(index) + row['legend']
    return (df, semsim_dict)


def scatterplot(df, output_dir_path, ontology, max_labels, palette, colour, size, legend, legend_position, max_clusters,
                legend_columns, cluster_labels, opacity, size_range, font_size, title, title_size, dpi, file_type,
                outfile_appendix, xlabel, ylabel, zlabel, logger_std):
    cmap = ListedColormap(sns.color_palette(palette))  ## Colour palette for clusters
    #cmap = sns.color_palette(palette, as_cmap=True)
    df = df.head(max_clusters)  ## If a maximum amount of clusters is given, truncate dataframe
    colour_criterium = df['colour']
    n = df.size
    if legend_columns == 'single':  ## Determine legend column number
        colnumber = 1
    elif legend_columns == 'double':
        colnumber = 2
    elif legend_columns == 'triple':
        colnumber = 3
    if size_range == 'small':  ## Determine size range of clusters
        size_range = (200, 2000)
    elif size_range == 'medium':
        size_range = (400, 4000)
    elif size_range == 'big':
        size_range = (600, 6000)
    if str.isdigit(str(size)):  ## If size range is changed using fixed integer, use that
        size = int(size) * 200
        size_range = (size, size)
    if colour.lower() == 'log10-pval' or colour.lower() == 'user' or colour.lower() == 'pval' or colour.lower() == 'members':  ## This and below if-statement to colour clusters
        norm = plt.Normalize(colour_criterium.min(), colour_criterium.max())
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        ax = sns.scatterplot(x='x', y='y', size='size', sizes=size_range, hue='colour', palette=cmap, edgecolor='black',
                             data=df.iloc[::-1], alpha=0.5, linewidth=1)
    elif colour.lower() == 'unique':
        ax = sns.scatterplot(x='x', y='y', size='size', sizes=size_range, hue='dotCount', palette=cmap,
                             edgecolor='black', data=df.iloc[::-1], alpha=0.5, linewidth=1)
    if legend_position.lower() == 'left':  ## This and below if-statement to set legend position
        legend_position = 'right'
    elif legend_position.lower() == 'right':
        legend_position = 'left'
    empty_handle = matplotlib.patches.Rectangle((0, 0), 1, 1, fill=False, edgecolor='none',
                                                visible=False)  ## Below statements fix the legend
    handle_list = [empty_handle] * max_labels
    legend_ax = ax.legend(handle_list, df.legend.iloc[0:max_labels], bbox_to_anchor=(0.5, -0.15),
                          loc='upper ' + legend_position.lower(), handlelength=0, handletextpad=0, ncol=colnumber,
                          frameon=False, fontsize=font_size)
    for item in legend_ax.legendHandles:
        item.set_visible(False)
    if legend == 'none':  ## No legend if excluded in options
        ax.get_legend().remove()

    if cluster_labels == 'numbered':  ## Couple of if-statements to determine how clusters are labeled
        if len(df.index) >= max_labels:
            for line in range(1, max_labels + 1):
                ax.text(df.x[line], df.y[line], str(line), horizontalalignment='center', verticalalignment='center',
                        size='small', color='black', weight='semibold', alpha=opacity)
        else:
            for line in range(1, len(df.index) + 1):
                ax.text(df.x[line], df.y[line], str(line), horizontalalignment='center', verticalalignment='center',
                        size='small', color='black', weight='semibold', alpha=opacity)
    elif cluster_labels == 'go' or cluster_labels == 'go-arrows' or cluster_labels == 'description' or cluster_labels == 'description-numbered':
        texts = []
        if cluster_labels == 'description-numbered':
            if len(df.index) >= max_labels:
                for line in range(1, max_labels + 1):
                    texts.append(ax.text(df.x[line], df.y[line], str(line) + '. ' + df.description[line],
                                         horizontalalignment='center', verticalalignment='center', size='small',
                                         color='black', weight='semibold', alpha=opacity))
            else:
                for line in range(1, len(df.index) + 1):
                    texts.append(ax.text(df.x[line], df.y[line], str(line) + '. ' + df.description[line],
                                         horizontalalignment='center', verticalalignment='center', size='small',
                                         color='black', weight='semibold', alpha=opacity))
        elif cluster_labels == 'description':
            if len(df.index) >= max_labels:
                for line in range(1, max_labels + 1):
                    texts.append(ax.text(df.x[line], df.y[line], df.description[line], horizontalalignment='center',
                                         verticalalignment='center', size='small', color='black', weight='semibold',
                                         alpha=opacity))
            else:
                for line in range(1, len(df.index) + 1):
                    texts.append(ax.text(df.x[line], df.y[line], df.description[line], horizontalalignment='center',
                                         verticalalignment='center', size='small', color='black', weight='semibold',
                                         alpha=opacity))
        else:
            if len(df.index) >= max_labels:
                for line in range(1, max_labels + 1):
                    texts.append(ax.text(df.x[line], df.y[line], str(line) + '. ' + df.representative[line],
                                         horizontalalignment='center', verticalalignment='center', size='small',
                                         color='black', weight='semibold', alpha=opacity))
            else:
                for line in range(1, len(df.index) + 1):
                    texts.append(ax.text(df.x[line], df.y[line], str(line) + '. ' + df.representative[line],
                                         horizontalalignment='center', verticalalignment='center', size='small',
                                         color='black', weight='semibold', alpha=opacity))

        if cluster_labels == 'go-arrows' or cluster_labels == 'description' or cluster_labels == 'description-numbered':
            adjust_text(texts, arrowprops=dict(arrowstyle='->', color='red',
                                               alpha=0.5))  ## Python package to make sure labels overlap as little as possible
        else:
            adjust_text(texts)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if not colour.lower() == 'unique':  ## Change label on the right based on values used for colour determination
        kwargs = {'format': '%.0f'}
        colorbar = ax.figure.colorbar(sm, alpha=0.5, **kwargs)
        if zlabel == None:
            if colour.lower() == 'user':
                colorbar.set_label(zlabel)

            elif colour.lower() == 'members':
                colorbar.set_label('Unique GO terms per cluster')
            elif colour.lower() == 'frequency':
                colorbar.set_label('Frequency of GO term in the GOA database')
            elif colour.lower() == 'user':
                colorbar.set_label('Significantly enriched GO terms per cluster')
            elif colour.lower() == 'pval':
                colorbar.set_label('pvalue')
            elif colour.lower() == 'log10-pval':
                colorbar.set_label('log10 pvalue')
        else:
            colorbar.set_label(zlabel)
        colorbar.ax.tick_params(size=0)
    if not title == None:
        ax.set_title(title, fontsize=title_size)
    fig = ax.get_figure()
    if outfile_appendix == '':
        outfile_name = ontology + '_' + output_dir_path.replace('/', '_') + '.' + file_type
    else:
        outfile_name = ontology + '_' + outfile_appendix + '.' + file_type
    fig.savefig(output_dir_path + '/' + outfile_name, dpi=dpi, bbox_extra_artists=(legend_ax,), bbox_inches='tight')
    plt.close()
    logger_std.info("Output for " + ontology.replace('_', ' ') + "found in: " + output_dir_path + "/" + outfile_name)


def output_table(df, output_dir_path, ontology, appendix):  ## Output raw table file based on the dataframe
    if appendix == '':
        df.to_csv(output_dir_path + '/' + ontology + '_' + output_dir_path.replace('/', '_') + '.tsv', sep='\t')
    else:
        df.to_csv(output_dir_path + '/' + ontology + '_' + appendix + '.tsv', sep='\t')


def create_cluster_table(semsim_dict, description_dict, output_dir_path, ontology, appendix):
    if appendix == '':
        outfile = open(output_dir_path + '/' + ontology + '_full_table_' + output_dir_path.replace('/', '_') + '.tsv',
                       'w')
    else:
        outfile = open(output_dir_path + '/' + ontology + '_full_table_' + appendix + '.tsv', 'w')
    outfile.write(
        'Cluster representative\tCluster member\tCluster member description\tMember P-value\tMember user defined value\tMember IC\tMember frequency')
    for representative, members in semsim_dict.items():
        member_set = set()
        for member in members:
            go, pval, user_val, ic, frequency = member
            description = description_dict[go]
            if not go in member_set:
                outfile.write('\n' + representative + '\t' + go + '\t' + description + '\t' + "{:.3e}".format(
                    pval) + '\t' + "{:.3e}".format(user_val) + '\t' + "{:.3e}".format(ic) + '\t' + "{:.3e}".format(
                    frequency))
            member_set.add(go)
    outfile.close()


"""
#!/usr/bin/env python3
import sys
from collections import defaultdict
from math import log10

goChildren = defaultdict(list)
goParents = defaultdict(list)
goCounts = defaultdict(int)
namespaceDict = defaultdict(str)

goSet = set()
for line in open(sys.argv[1]): ## GO parents
	if not line.startswith('#'):
		go,parent = line.strip().split('\t')
		goSet.add(go)
		goSet.add(parent)
		goParents[go].append(parent)

for line in open(sys.argv[1]): ## GO children
	if not line.startswith('#'):
		child,go = line.strip().split('\t')
		goChildren[go].append(child)

for line in open(sys.argv[3]):
	if line.startswith('id: GO:'):
		go = line.strip().split('id: ')[1]
	elif line.startswith('namespace: '):
		namespace = line.strip().split('namespace: ')[1]
		namespaceDict[go] = namespace

totMF = 0
totBP = 0
totCC = 0
for line in open(sys.argv[2]):
	if not line.startswith('!'):
		ssline = line.strip().split('\t')
		go = ssline[4]
		goCounts[go] += 1
	else:
		if line.startswith('!Generated: '):
			goa_version = line.strip().split(' ')[1]
			print('# GOA version used: '+goa_version)

for go,count in goCounts.items():
	if namespaceDict[go] == 'biological_process':
		totBP += count
	elif namespaceDict[go] == 'molecular_function':
		totMF += count
	elif namespaceDict[go] == 'cellular_component':
		totCC += count

goCountsDownpropagated = defaultdict(int)
for go in goSet:
	counts = goCounts[go]
	goCountsDownpropagated[go] = counts
	for child in goChildren[go]:
		goCountsDownpropagated[go] += goCounts[child]

print('#GO\tIC\tFrequency')
for go in goCountsDownpropagated:
	if namespaceDict[go] == 'biological_process':
		tot = totBP
	elif namespaceDict[go] == 'molecular_function':
		tot = totMF
	elif namespaceDict[go] == 'cellular_component':
		tot = totCC
	ic = 0
	frequency = 0
	if not goCountsDownpropagated[go] == 0:
		ic = -log10(goCountsDownpropagated[go]/tot)
		frequency = goCountsDownpropagated[go]/tot
	else:
		ic = -log10(1/tot)
		frequency = 1/tot
	print(go+'\t'+str(ic)+'\t'+str(frequency))
"""


import re
import os
import json
import h5py
import anndata
import h5sparse
import scipy.io
import numpy as np
import pandas as pd

from pathlib import Path
from anndata import AnnData
from matplotlib.image import imread
from scipy.sparse import csr_matrix, csc_matrix
from sklearn.preprocessing import normalize as sp_normalize

from ._utilities import _h5_idx_generator, _create_h5_using_adata


def read_10X(sample_dirs,
             sample_names,
             merge=True,
             num_cells=None,
             min_umis=0,
             use_filtered=False,
             reference=None,
             data_type="rna",
             verbose=True):
    """Read 10X alignment data (including V3)

    This function generates a sparse matrix (cells x genes) from the data generated by 10X's
    cellranger count pipeline. It can process V2 and V3 data together, producing either a single
    merged matrix or list of matrices. Also handles multiple data types produced by 10X V3 (Gene
    Expression, Antibody Capture, CRISPR, CUSTOM).

    Parameters
    ----------
    sample_dirs : list
        List of directories containing either matrix.mtx(.gz) file along with genes.tsv,
        (features.tsv), and barcodes.tsv, or outer level 10X output directory (containing outs directory).
    sample_names : list
        List of names to use for samples (corresponding to sample_dirs)
    merge : bool, optional
        Whether to merge all matrices of the same data type across samples or leave as list
        of matrices (the default is True).
    num_cells : int, optional
        Optional limit on number of cells returned for each sample (only for Gene
        Expression data). Retains the cells with the highest numbers of transcripts
        (the default is None).
    min_umis : int, optional
        Minimum UMI threshold for cells (the default is 0).
    use_filtered : bool, optional
        Whether to use 10X's filtered data (as opposed to raw). Only relevant for
        sample.dirs containing 10X outs directory (the default is False).
    reference : str, optional
        For 10X V<3, specify which reference directory to use if sample_dir is outer
        level 10X directory (only necessary if more than one reference used for sequencing)
        (the default is None).
    data_type : str, optional, 'rna' or 'atac'
        Indicates the protocol of the input data. If not specified, input data will be
        considered scRNA-seq data (the default is 'rna').
    verbose : bool, optional
        Print messages (True by default)

    Returns
    -------
    datalist : list
         List of merged matrices stored as AnnData objects across data types
         (returns sparse matrix if only one data type detected), or nested
         list of matrices organized by sample if merge=F.

    Examples
    --------
    >>> sample_dir1 = "path/to/outer/dir1" # 10X output directory V2 -- contains outs/raw_gene_bc_matrices/<reference>/...
    >>> sample_dir2 = "path/to/outer/dir2" # 10X output directory V3 -- for two data types, Gene Expression and CUSTOM
    >>> dges1 = read_10x([sample_dir1, sample_dir2], ["sample1", "sample2"], min_umis = 50)
    >>> ligerex = create_liger(expr = dges1[["Gene Expression"]], custom = dges1[["CUSTOM"]])
    """

    data_list = []
    data_types = np.array(['Gene Expression'])
    num_samples = len(sample_dirs)

    if num_cells is not None:
        num_cells = np.repeat(num_cells, num_samples)

    for i in range(num_samples):
        # start message
        print('Processing sample ' + sample_names[i])

        # Construct sample path
        sample_dir, suffix = _build_path(sample_dirs[i], use_filtered, reference)

        matrix_file = sample_dir + '/matrix.mtx' + suffix
        barcodes_file = sample_dir + "/barcodes.tsv" + suffix

        # Read in raw data (count matrix)
        raw_data = scipy.io.mmread(matrix_file)

        # filter for UMIs first to increase speed
        umi_pass = np.sum(raw_data, axis=0) > min_umis
        umi_pass = np.asarray(umi_pass).flatten()  # convert to np array
        if umi_pass.shape[0] == 0:
            print('No cells pass UMI cutoff. Please lower it.')
        raw_data = raw_data[:, umi_pass]
        raw_data = csr_matrix(raw_data)  # convert to csr matrix

        # Create column names
        barcodes = pd.read_csv(barcodes_file, sep='\t', header=None)
        barcodes = barcodes.to_numpy().flatten()[umi_pass]

        # remove -1 tag from barcodes
        for i in range(barcodes.size):
            barcodes[i] = re.sub('\-1$', '', barcodes[i])

        col_names = pd.DataFrame(barcodes, columns=['barcodes'])

        # Create row names
        if data_type == 'rna':
            features = pd.read_csv(features_file, sep='\t', header=None).to_numpy()  # convert to np array
            row_names = features[:, 1]

            # equal to make.unique function in R
            count_dict = {}
            for i in range(len(row_names)):
                name = row_names[i]
                if name not in count_dict:
                    count_dict[name] = 0
                if name in row_names:
                    count_dict[name] += 1
                    if count_dict[name] > 1:
                        row_names[i] = row_names[i] + '.' + str(count_dict[name] - 1)

            row_names = pd.DataFrame(row_names, columns=['gene_name'])

        elif data_type == 'atac':
            features = pd.read_csv(features_file, sep='\t', header=None).to_numpy()
            features = np.array(
                [str(feature[0]) + ':' + str(feature[1]) + '-' + str(feature[2]) for feature in features])
            row_names = pd.DataFrame(features, columns=['gene_name'])

        # split based on 10X datatype -- V3 has Gene Expression, Antibody Capture, CRISPR, CUSTOM
        # V2 has only Gene Expression by default and just two columns
        # TODO: check atac feature file
        if features.shape[1] == 1:
            sample_datatypes = np.array(['Chromatin Accessibility'])
            adata = AnnData(csr_matrix(raw_data), obs=row_names, var=col_names)
            adata.uns['sample_name'] = sample_names[i]
            adata.uns['data_type'] = 'Chromatin Accessibility'
            data_list.append(adata)
        elif features.shape[1] < 3:
            sample_datatypes = np.array(['Gene Expression'])
            adata = AnnData(csr_matrix(raw_data), obs=row_names, var=col_names)
            adata.uns['sample_name'] = sample_names[i]
            adata.uns['data_type'] = 'Gene Expression'
            data_list.append(adata)
        else:
            sample_datatypes = features[:, 2]
            sample_datatypes_unique = np.unique(sample_datatypes)
            # keep track of all unique datatypes
            data_types = np.union1d(data_types, sample_datatypes_unique)

            for name in sample_datatypes:
                idx = sample_datatypes == name
                subset_row_names = row_names[idx]
                subset_row_names = pd.DataFrame(subset_row_names, columns=['gene_name'])
                subset_data = raw_data[:, sample_datatypes == name]
                adata = AnnData(csr_matrix(subset_data), obs=subset_row_names, var=col_names)
                adata.uns['sample_name'] = sample_names[i]
                adata.uns['data_type'] = name
                data_list.append(adata)

        # num_cells filter only for gene expression data
    #        if num_cells is not None:
    #            if 'Gene Expression' or 'Chromatin Accessibility' in sample_datatypes and sample_datatypes.shape[0] == 1:
    #                data_label = sample_datatypes.item()
    #                cs = np.sum(samplelist[data_label], axis=0)
    #                limit = samplelist[data_label].shape[1]
    #                if num_cells[i] > limit:
    #                    print('You selected more cells than are in matrix {}. Returning all {} cells.'.format(i, limit))
    #                num_cells[i] = limit
    #                samplelist[data_label] = np.flip(np.sort(samplelist[data_label]))[0:num_cells[i]]

    return_dges = {}
    for datatype in data_types:
        for data in data_list:
            if datatype not in return_dges:
                return_dges[datatype] = []
            else:
                return_dges[datatype].append(data[data])

        return_dges.append(adata)
    if merge:
        print('Merging samples')

        # return_dges = MergeSparseDataAll()
        # if only one type of data present
        if len(return_dges) == 1:
            print('Returning {} data matrix'.format(datatypes))

    else:
        return data_list

    return data_list


def read_10X_h5(sample_dir: str,
                sample_name: str,
                file_name = None,
                use_raw = False,
                backed = False,
                chunk_size = 1000
                ) -> AnnData:
    """
    need to be improved a lot
    :param sample_dir:
    :param sample_name:
    :return:
    """
    ### 0. Parameter checking
    if file_name is not None:
        file_path = Path(sample_dir, file_name)
    else:
        if use_raw:
            file_path = Path(sample_dir, 'raw_feature_bc_matrix.h5')
        else:
            file_path = Path(sample_dir, 'filtered_feature_bc_matrix.h5')

    ### 1. Read h5 file
    with h5py.File(file_path, 'r') as f:
        raw_data = csc_matrix((f['matrix']['data'], f['matrix']['indices'], f['matrix']['indptr']),
                              shape=f['matrix']['shape'])
        features = pd.DataFrame({'gene_name': f['matrix']['features']['name'][()].astype(str)})
        barcodes = pd.DataFrame({'barcodes': f['matrix']['barcodes'][()].astype(str)})

    raw_data = csr_matrix(raw_data.transpose())
    barcodes = barcodes.set_index('barcodes')
    features = features.set_index('gene_name')

    ### 2. Save AnnData or hdf5 if backed
    if not backed:
        adata = AnnData(raw_data, obs=barcodes, var=features,
                        uns={'sample_name': sample_name, 'data_type': 'Gene Expression'})
    else:
        if not os.path.isdir('./results'):
            os.mkdir('./results')

        """
        # create h5 file for each individual sample.
        file_name = './results/' + sample_name + '.hdf5'
        with h5sparse.File(file_name, 'w') as f:
            for left, right in _h5_idx_generator(chunk_size, raw_data.shape[0]):
                if 'raw_data' not in f.keys():
                    f.create_dataset('raw_data', data=raw_data[left:right, :], chunks=(chunk_size,), maxshape=(None,))
                else:
                    f['raw_data'].append(raw_data[left:right, :])
        """
        adata = AnnData(raw_data, obs=barcodes, var=features,
                        uns={'sample_name': sample_name, 'data_type': 'Gene Expression', 'backed_path': file_name})
        _create_h5_using_adata(adata, chunk_size)
        adata.write_h5ad(filename='./results/' + sample_name + '.h5ad')
        adata = anndata.read_h5ad(filename='./results/' + sample_name + '.h5ad', backed=True)

    adata.var_names_make_unique()

    return adata




def read_10X_visium(sample_dirs, sample_names):
    """

    :param sample_dirs:
    :param sample_names:
    :return:
    """
    ### 1. read count matrix
    adata = read_10X_h5(sample_dirs, sample_names)

    ### 2. read spatial info
    p_spatial = Path(sample_dirs, 'spatial')

    if not p_spatial.is_dir():
        print('Spatial file is not existed.')
    else:
        positions = pd.read_csv((p_spatial / 'tissue_positions_list.csv'), header=None)
        positions.columns = [
            'barcodes',
            'in_tissue',
            'array_row',
            'array_col',
            'pxl_row_in_fullres',
            'pxl_col_in_fullres'
        ]
        positions = positions.set_index('barcodes')

        adata.obs = adata.obs.join(positions, how='left')
        adata.obsm['pxl_in_fullres'] = adata.obs[['pxl_row_in_fullres', 'pxl_col_in_fullres']].to_numpy()
        adata.obsm['array_spatial'] = adata.obs[['array_row', 'array_col']].to_numpy()
        adata.obs.drop(
            columns=['array_row', 'array_col', 'pxl_row_in_fullres', 'pxl_col_in_fullres'],
            inplace=True,
        )

        adata.uns['image'] = {
            'lowres': imread(p_spatial / 'tissue_lowres_image.png'),
            'hires': imread(p_spatial / 'tissue_hires_image.png'),
            'scale_factors': json.loads((p_spatial / 'scalefactors_json.json').read_text())
        }

    return adata


def read_10X_multiome(sample_dir, sample_name):
    """

    :param sample_dir:
    :param sample_name:
    :return:
    """
    file_name = sample_dir + '/' + sample_name + '.h5'

    with h5py.File(file_name, 'r') as f:
        raw_data = csc_matrix((f['matrix']['data'], f['matrix']['indices'], f['matrix']['indptr']),
                              shape=f['matrix']['shape'])
        features = pd.DataFrame({'gene_name': f['matrix']['features']['name'][()].astype(str),
                                 'id': f['matrix']['features']['id'][()].astype(str),
                                 'feature_type': f['matrix']['features']['feature_type'][()].astype(str)
                                 })
        barcodes = pd.DataFrame({'barcodes': f['matrix']['barcodes'][()].astype(str)})

    raw_data = csr_matrix(raw_data.transpose())
    barcodes = barcodes.set_index('barcodes')
    features = features.set_index('gene_name')

    rna_idx = features['feature_type'] == 'Gene Expression'
    adata_atac = AnnData(raw_data[:, ~rna_idx], obs=barcodes, var=features[~rna_idx],
                         uns={'sample_name': sample_name, 'data_type': 'Gene Expression'})
    adata_rna = AnnData(raw_data[:, rna_idx], obs=barcodes, var=features[rna_idx],
                        uns={'sample_name': sample_name, 'data_type': 'Gene Expression'})
    adata_atac.var_names_make_unique()
    adata_rna.var_names_make_unique()

    return adata_atac, adata_rna


def read_10X_atac():
    pass

def save(dir, overwrite='None'):
    p = Path(dir)
    if p.exists():
        if overwrite is True:
            pass
    pass


def save_cellxgene(liger_object,
                   save_dir,
                   return_adata=False):
    """ """
    concat_adata = anndata.concat(liger_object.adata_list)
    cellxgene_adata = AnnData(csc_matrix(np.log2(sp_normalize(concat_adata.raw.X, axis=1, norm='l1').toarray() * 10000 + 1)),
                              obs=concat_adata.obs, var=concat_adata.raw.var, uns=concat_adata.uns, obsm=concat_adata.obsm)

    cellxgene_adata.obsm['X_umap'] = cellxgene_adata.obsm['umap_coords']
    cellxgene_adata.obs['leiden'] = cellxgene_adata.obs['cluster'].astype("category")


    if return_adata:
        return cellxgene_adata
    else:
        return None


def write_h5(liger_object,
             save_dir):
    """helper function to save a copy of h5 file"""
    pass


def load(dir, backed='None'):
    pass

def _read_mtx(sample_dirs):
    pass
    #return adata

def _read_10x_v3_h5(sample_dirs):
    pass
    #return adata


def _build_path(sample_dir, use_filtered, reference):
    check_inner = sample_dir + '/outs'

    if os.path.exists(check_inner):
        sample_dir = check_inner
        is_v3 = os.path.exists(sample_dir + '/filtered_feature_bc_matrix')
        matrix_prefix = str(np.where(use_filtered, 'filtered', 'raw'))

        if is_v3:
            sample_dir = sample_dir + '/' + matrix_prefix + '_feature_bc_matrix'
        else:
            if reference is None:
                references = os.listdir(sample_dir + '/raw_gene_bc_matrices')
                if len(references) > 1:
                    raise ValueError('Multiple reference genomes found. Please specify a single one.')
                else:
                    reference = references[0]
        if reference is None:
            reference = ''
        sample_dir = sample_dir + '/' + matrix_prefix + '_gene_bc_matrices/' + reference
    else:
        is_v3 = os.path.exists(sample_dir + '/features.tsv.gz')

    suffix = str(np.where(is_v3, '.gz', ''))
    if data_type == 'rna':
        features_file = str(np.where(is_v3, sample_dir + '/features.tsv.gz', sample_dir + '/genes.tsv'))
    elif data_type == 'atac':
        features_file = str(np.where(is_v3, sample_dir + '/peaks.bed.gz', sample_dir + '/peaks.bed'))

    return sample_dir, suffix

class Data:
    def __init__(self, file_path, file_name):
        pass


import numpy as np
import pandas as pd
import matplotlib as mpl
from tqdm import tqdm
from plotnine import *

from ..tools import calc_dataset_specificity, get_factor_markers


def plot_gene_loadings(liger_object,
                       dataset1 = None,
                       dataset2 = None,
                       num_genes_show = 12,
                       num_genes = 30,
                       mark_top_genes = True,
                       factor_share_thresh = 10,
                       log_fc_thresh = 1,
                       umi_thresh = 30,
                       frac_thresh = 0,
                       pval_thresh = 0.05,
                       do_spec_plot = True,
                       max_val = 0.1,
                       pt_size = 0.1,
                       option = 'plasma_r',
                       zero_color = "#F5F5F5",
                       return_plots = False,
                       axis_labels = None,
                       do_title = False):
    """Generate t-SNE plots and gene loading plots

    Plots t-SNE coordinates of all cells by their loadings on each factor. Underneath it displays the
    most highly loading shared and dataset-specific genes, along with the overall gene loadings
    for each dataset.

    It is recommended to call this function into a PDF due to the large number of
    plots produced.

    """
    ### Parameter setting
    if dataset1 is None or dataset2 is None:
        #dataset1 = liger_object.adata_list[0].uns['sample_name']
        #dataset2 = liger_object.adata_list[1].uns['sample_name']
        dataset1 = 0
        dataset2 = 1

    ### Extract Values
    H_aligned = liger_object.get_obsm('H_norm')
    W_orig = liger_object.W
    V1 = liger_object.V[dataset1]
    V2 = liger_object.V[dataset2]
    W = np.minimum(W_orig+V1, W_orig+V2)

    dataset_specificity = calc_dataset_specificity(liger_object, dataset1=dataset1, dataset2=dataset2, do_plot=do_spec_plot)

    factors_use = np.abs(dataset_specificity[2][dataset_specificity[2] <= factor_share_thresh])

    markers = get_factor_markers(liger_object, dataset1=dataset1, dataset2=dataset2,
                                 factor_share_thresh=factor_share_thresh,
                                 num_genes=num_genes, log_fc_thresh=log_fc_thresh, umi_thresh=umi_thresh,
                                 frac_thresh=frac_thresh, pval_thresh=pval_thresh,
                                 dataset_specificity=dataset_specificity)

    loadings_list = [V1, W, V2]

    names_list = [dataset1, 'Shared', dataset2]
    tsne_coords = liger_object.tsne_coords
    return_plots = []

    ### Create plot
    for i in tqdm(factors_use):
        factorlab = ('Factor ' + str(i))
        tsne_df = tsne_coords.append({factorlab: H_aligned[:, i]})
        tsne_df[tsne_df[factorlab] == 0][factorlab] = np.nan
        factor_ds = (factorlab + 'Dataset Specificity: ' + dataset_specificity[2][i])
        data_max = np.max(H_aligned[:, i])

        # plot t-SNE
        if max_val is not None:
            values = [0, max_val, 1]
        else:
            values = None

        def _rescale(x, _from):
            norm = mpl.colors.TwoSlopeNorm(vmin=_from[0], vcenter=_from[0] + max_val*(_from[1]-_from[0]), vmax=_from[1])
            return norm(x)

        p1 = (ggplot(data=tsne_df, mapping=aes(x='tsne1', y='tsne2', color=factorlab)) +
              geom_point(size=pt_size) +
              scale_color_cmap(cmap_name=option, na_value=zero_color, rescaler=_rescale) +
              theme_classic(12))

        if axis_labels is not None:
            p1 = p1 + xlab(axis_labels[0]) + ylab(axis_labels[1])

        if do_title:
            p1 = p1 + ggtitle(factor_ds)

        # subset to specific factor and sort by p-value
        top_genes_V1 = markers[0][markers[0]['factor_num'] == i]
        top_genes_V1 = top_genes_V1.sort_values(by=['p_value'])['gene']

        # don't sort for W
        top_genes_W = markers[1][markers[1]['factor_num'] == i]['gene']
        top_genes_V2 = markers[2][markers[2]['factor_num'] == i]
        top_genes_V2 = top_genes_V2.sort_values(by=['p_value'])['gene']

        top_genes_list = [top_genes_V1, top_genes_W, top_genes_V2]
        plot_list = []
        for idx, gene_list in enumerate(top_genes_list):
            # subset down to those which will be shown if sorting by p-val
            if len(gene_list) > num_genes_show:
                top_genes_list[idx] = gene_list[1:num_genes_show]

            top_genes = top_genes_list[idx]

            # make dataframe for cum gene loadings plot
            sorted = np.argsort(loadings_list[idx][:, i])

            # sort by loadings instead - still only showing num_genes_show
            # look through top num.genes in loadings
            top_loaded = sorted[len(sorted):(len(sorted)-num_genes):-1]
            top_genes = top_genes[top_loaded]
            if len(top_genes) == 0:
                top_genes = ['no genes']

            gene_df = pd.DataFrame({'loadings': sorted,
                                    'xpos': np.linspace(0,1,num=len(sorted)+1),
                                    'top_k': 0})
            y_lim_text = gene_df['loadings'].max()

            # plot and annotate with top genes
            out_plot = (ggplot(data=gene_df, mapping=aes(x='xpos', y='loadings')) +
                        geom_point(size=0.4) +
                        theme_bw() +
                        theme(axis_ticks_direction_x=element_blank(),
                              axis_line_x=element_blank(),
                              axis_title=element_blank(),
                              axis_text_x=element_blank(),
                              panel_grid_major_x=element_blank(),
                              panel_grid_minor_x=element_blank()) +
                        ggtitle(names_list[idx]) +
                        annotate('text', x=1.1,
                                 y=np.linspace(y_lim_text, 0 ,num=num_genes_show),
                                 label=top_genes,
                                 hjust=0,
                                 col='#8227A0') +
                        coord_cartesian(xlim=[0,1], # this focuses the x-axis on the range of interest
                                        clip='off') +
                        theme()) #TODO
            if mark_top_genes:
                out_plot = out_plot + geom_point(data=gene_df, mapping=aes('xpos', 'loadings'), color='#8227A0', size=0.5)

            plot_list.append(out_plot)


        if not return_plots:
            return_plots[i].draw()

    if return_plots:
        return return_plots



from ._metrics import calc_dataset_specificity
from ._umap import run_umap
from ._marker import get_factor_markers
from ._wilcoxon import run_wilcoxon, _wilcoxon
from ._quantile_norm import quantile_norm
from ._go import run_GO_analysis

#######################################################################################
#### Conversion/Transformation TODO

# Create a Seurat object containing the data from a liger object
# TO-DO names function
# def ligerToSeurat(liger_object, nms = names(object@H), renormalize = True, use_liger_genes = True,
#                  by_dataset = False):
#    pass

# Create liger object from one or more Seurat objects
def seuratToLiger(liger_object, combined_seurat=False, names="use-projects", meta_var=None,
                  assays_use=None, raw_assay="RNA", remove_missing=True, renormalize=True,
                  use_seurat_genes=True, num_hvg_info=None, use_idents=True, use_tsne=True,
                  cca_to_H=False):
    pass


# Construct a liger object with a specified subset
def subsetLiger(liger_object, clusters_use=None, cells_use=None, remove_missing=True):
    pass


# Construct a liger object organized by another feature
def reorganizeLiger(liger_object, by_feature, keep_meta=True, new_label="orig.dataset"):
    pass


# Convert older liger object into most current version (based on class definition)
def convertOldLiger(liger_object, override_raw=False):
    pass



from .read_write import read_10X, read_10X_h5, read_10X_visium
from .preprocessing import make_feature_matrix, create_liger, normalize, select_genes, scale_not_center
from .factorization import optimize_ALS, iNMF_HALS, online_iNMF
from .clustering import louvain_cluster, leiden_cluster
from .plotting import plot_by_dataset_and_cluster, plot_gene, plot_gene_loadings, plot_spatial
from .tools import quantile_norm, run_umap, run_wilcoxon




#from ._qc import plot_qc
from ._gene import plot_gene
from ._gene_loadings import plot_gene_loadings
from ._dataset_cluster import plot_by_dataset_and_cluster
from ._spatial import plot_spatial
from ._go import plot_go_term


import numpy as np
###############TODO
# Plot specific feature on t-SNE coordinates
def plotFeature(liger_object, feature, by_dataset=True, discrete=None, title=None,
                pt_size=0.3, text_size=3, do_shuffle=True, rand_seed=1, do_labels=False,
                axis_labels=None, do_legend=True, legend_size=5, option='plasma',
                cols_use=None, zero_color='#F5F5F5', return_plots=False):
    pass


# Plot scatter plots of unaligned and aligned factor loadings
def plotFactors(liger_object, num_genes=10, cells_highlight=None, plot_tsne=False):
    pass


# Generate word clouds and t-SNE plots
def plotWordClouds(liger_object, dataset1=None, dataset2=None, num_genes=30, min_size=1,
                   max_size=4, factor_share_thresh=10, log_fc_thresh=1,
                   umi_thresh=30, frac_thresh=0, pval_thresh=0.05,
                   do_spec_plot=True, return_plots=False):
    pass


# Plot violin plots for gene expression
def plotGeneViolin(liger_object, gene, methylation_indices=None,
                   by_dataset=True, return_plots=False):
    pass


# Plot expression of multiple genes
def plotGenes(liger_object, genes):
    pass


# Generate a river (Sankey) plot
def makeRiverplot(liger_object, cluster1, cluster2, cluster_consensus=None, min_frac=0.05,
                  min_cells=10, river_yscale=1, river_lty=0, river_node_margin=0.1,
                  label_cex=1, label_col="black", lab_srt=0, river_usr=None,
                  node_order="auto"):
    pass


# ' Plot cluster proportions by dataset
def plotClusterProportions(liger_object, return_plot=False):
    pass


# Plot heatmap of cluster/factor correspondence
def plotClusterFactors(liger_object, use_aligned=False, Rowv=np.nan, Colv="Rowv", col=None,
                       return_data=False):
    pass


import louvain
import numpy as np

from ._utilities import run_knn, compute_snn, build_igraph, _assign_cluster


def louvain_cluster(liger_object,
                    resolution=1.0,
                    k=20,
                    prune=1 / 15,
                    random_seed=1,
                    n_starts=10):
    """Louvain algorithm for community detection

    After quantile normalization, users can additionally run the Louvain algorithm
    for community detection, which is widely used in single-cell analysis and excels at merging
    small clusters
    nto broad cell classes.
    Parameters
    ----------
    liger_object : liger object
        Should run quantile_norm before calling.
    resolution : float, optional
        Value of the resolution parameter, use a value above (below) 1.0 if you want
        to obtain a larger (smaller) number of communities (the default is 1.0).
    k : int, optional
        The maximum number of nearest neighbours to compute (the default is 20).
    prune : float, optional
        Sets the cutoff for acceptable Jaccard index when
        computing the neighborhood overlap for the SNN construction. Any edges with
        values less than or equal to this will be set to 0 and removed from the SNN
        graph. Essentially sets the strigency of pruning (0 --- no pruning, 1 ---
        prune everything) (the default is 1/15).
    random_seed : int, optional
        Seed of the random number generator (the default is 1).
    n_starts : int, optional
        The number of random starts to be used (the default is 10).
    Returns
    -------
    liger_object : liger object
        object with refined 'cluster'.

    Examples
    --------
    >>> ligerex = louvain_cluster(ligerex, resulotion = 0.3) # liger object, factorization complete
    """
    ### 1. Compute snn
    H_norm = np.vstack([adata.obsm['H_norm'] for adata in liger_object.adata_list])
    knn = run_knn(H_norm, k)
    snn = compute_snn(knn, prune=prune)

    ### 2. Get igraph from snn
    g = build_igraph(snn)

    ### 3. Run louvain
    np.random.seed(random_seed)
    max_quality = -1
    for i in range(n_starts):  # random starts to improve stability
        seed = np.random.randint(0, 1000)
        kwargs = {'weights': g.es['weight'], 'resolution_parameter': resolution, 'seed': seed}  # parameters setting
        part = louvain.find_partition(g, louvain.RBConfigurationVertexPartition, **kwargs)

        if part.quality() > max_quality:
            cluster = part.membership
            max_quality = part.quality()

    ### 4. Assign cluster results
    _assign_cluster(cluster, liger_object)

    return None
